import logging

from forge.mcrypto.hasher import Hasher
from forge.mcrypto.signer import Signer
from forge.utils import utils

logger = logging.getLogger('abt-did')


class AbtDid:
    PREFIX = 'did:abt:'
    MIN_30 = 1800
    ROLE_MAP = {
        'account': 0,
        'node': 1,
        'device': 2,
        'application': 3,
        'smart_contract': 4,
        'bot': 5,
        'asset': 6,
        'stake': 7,
        'validator': 8,
        'group': 9,
        'any': 63,
    }

    KEY_MAP = {
        'ed25519': 0,
        'secp256k1': 1,
    }

    HASH_MAP = {
        'keecak': 0,
        'sha3': 1,
        'keccak_384': 2,
        'sha3_384': 3,
        'keccak_512': 4,
        'sha3_512': 5,
    }

    def __init__(self, role_type='account',
                 key_type='ed25519',
                 hash_type='sha3', **kwargs):
        self.role_type = role_type
        assert (role_type in AbtDid.ROLE_MAP.keys())

        self.key_type = key_type
        assert (key_type in AbtDid.KEY_MAP.keys())
        self.signer = Signer(key_type)

        self.hash_type = hash_type
        assert (hash_type in AbtDid.HASH_MAP.keys())
        self.hasher = Hasher(hash_type)

        self.encode = kwargs.get('encode', True)
        self.form = kwargs.get('form', 'long')

    @staticmethod
    def parse_type_from_did(did):
        try:
            did = did.lstrip(AbtDid.PREFIX)
            decoded = utils.multibase_b58decode(did)
            type_bytes = decoded[0:2]
            return AbtDid.bytes_to_type(type_bytes)
        except Exception as e:
            logger.error('Fail to parse type from given did {}'.format(did))
            logger.error(e, exc_info=True)

    def sk_to_did(self, sk):
        pk = self.signer.sk_to_pk(sk)
        return self.pk_to_did(pk)

    def pk_to_did(self, pk):
        type_bytes = self.type_to_bytes()
        pk_hash = self.hasher.hash(pk)
        extended_hash = type_bytes + pk_hash[0:20]

        did_bytes = extended_hash + self.hasher.hash(extended_hash)[0:4]
        encoded_did = utils.multibase_b58encode(did_bytes)

        if not self.encode:
            return did_bytes
        elif self.form == 'long':
            return AbtDid.PREFIX + encoded_did
        else:
            return encoded_did

    @staticmethod
    def is_match_pk(did, pk):
        if did.startswith(AbtDid.PREFIX):
            did = did.lstrip(AbtDid.PREFIX)
        try:
            decoded = utils.multibase_b58decode(did)
            type_bytes = decoded[0:2]
            did_type = AbtDid.bytes_to_type(type_bytes)
            if did == did_type.pk_to_did(pk).lstrip(AbtDid.PREFIX):
                return True
            return False
        except Exception as e:
            logger.error('Fail to match pk {}'.format(pk))
            logger.error(e, exc_info=True)
            return False

    @staticmethod
    def is_valid(did):
        did = did.lstrip('did:abt:')
        try:
            decoded = utils.multibase_b58decode(did)
            type_bytes = decoded[0:2]
            pk_hash = decoded[2:22]
            actual_check_sum = decoded[22:26]

            did_type = AbtDid.bytes_to_type(type_bytes)
            expectued_check_sum = did_type.hasher.hash(type_bytes + pk_hash)[
                0:4]
            if actual_check_sum == expectued_check_sum:
                return True
            return False
        except Exception as e:
            logger.error('Fail to verify did {}'.format(did))
            logger.error(e, exc_info=True)
            return False

    def type_to_bytes(self):
        role_bits = to_six_bits(AbtDid.ROLE_MAP.get(self.role_type))
        key_bits = to_five_bits(AbtDid.KEY_MAP.get(self.key_type))
        hash_bits = to_five_bits(AbtDid.HASH_MAP.get(self.hash_type))

        first_byte = bytes([int((role_bits + key_bits[0:2]), 2)])
        second_byte = bytes([int((key_bits[2:] + hash_bits), 2)])
        return first_byte + second_byte

    @staticmethod
    def bytes_to_type(input_bytes):
        bits = bin(utils.bytes_to_int(input_bytes) | 65536)[3:]
        role_type = get_did_type_key(AbtDid.ROLE_MAP, bits[0:6])
        key_type = get_did_type_key(AbtDid.KEY_MAP, bits[6:11])
        hash_type = get_did_type_key(AbtDid.HASH_MAP, bits[11:16])

        return AbtDid(role_type=role_type,
                      key_type=key_type,
                      hash_type=hash_type)

    def gen_and_sign(self, sk, extra):
        now = utils.current_utc_timestamp()
        middle = utils.clean_dict({'iss': self.sk_to_did(sk),
                                   'iat': now,
                                   'nbf': now,
                                   'exp': now + AbtDid.MIN_30,
                                   **extra,
                                   })
        body = utils.to_json_b64urlsafe(middle)
        data = self.__header() + '.' + body
        signature = utils.b64encode_no_padding(
            self.signer.sign(data.encode(), sk))
        return data + '.' + signature

    def __header(self):
        if self.key_type == 'ed25519':
            alg = 'Ed25519'
        elif self.key_type == 'secp256k1':
            alg = 'ES256K'

        return utils.to_json_b64urlsafe({
            'alg': alg,
            'typ': 'JWT'
        })

    @staticmethod
    def verify(token, pk):
        try:
            header, body, signature = token.split('.')
            alg = utils.b64encoded_to_dict(header).get('alg').lower()
            if alg == 'secp256k1' or alg == 'es256k':
                signer = Signer('secp256k1')
            elif alg == 'ed25519':
                signer = Signer('ed25519')

            sig = utils.b64decode_padding_safe(signature)
            is_sig_valid = signer.verify((header + '.' + body).encode(),
                                         sig, pk)
            did = utils.b64encoded_to_dict(body).get('iss')
            if is_sig_valid and AbtDid.is_match_pk(did, pk):
                return True
            return False
        except Exception as e:
            logger.error(e, exc_info=True)
            logger.error("Fail to verify token {0} and pk {1}"
                         .format(token, pk))
            return False


def get_did_type_key(dic, bits):
    for k, v in dic.items():
        if v == int(b'0b' + bits.encode(), 2):
            return k
    raise ValueError('Bits {} does not have a corresponded type.'.format(bits))


def to_six_bits(num):
    return bin(num | 64)[3:]


def to_five_bits(num):
    return bin(num | 32)[3:]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import vendor_pb2 as vendor__pb2


class ABCIApplicationStub(object):
    """----------------------------------------
    Service Definition

    """

    def __init__(self, channel):
        """Constructor.

        Args:
          channel: A grpc.Channel.
        """
        self.Echo = channel.unary_unary(
            '/abci_vendor.ABCIApplication/Echo',
            request_serializer=vendor__pb2.RequestEcho.SerializeToString,
            response_deserializer=vendor__pb2.ResponseEcho.FromString,
        )
        self.Flush = channel.unary_unary(
            '/abci_vendor.ABCIApplication/Flush',
            request_serializer=vendor__pb2.RequestFlush.SerializeToString,
            response_deserializer=vendor__pb2.ResponseFlush.FromString,
        )
        self.Info = channel.unary_unary(
            '/abci_vendor.ABCIApplication/Info',
            request_serializer=vendor__pb2.RequestInfo.SerializeToString,
            response_deserializer=vendor__pb2.ResponseInfo.FromString,
        )
        self.SetOption = channel.unary_unary(
            '/abci_vendor.ABCIApplication/SetOption',
            request_serializer=vendor__pb2.RequestSetOption.SerializeToString,
            response_deserializer=vendor__pb2.ResponseSetOption.FromString,
        )
        self.DeliverTx = channel.unary_unary(
            '/abci_vendor.ABCIApplication/DeliverTx',
            request_serializer=vendor__pb2.RequestDeliverTx.SerializeToString,
            response_deserializer=vendor__pb2.ResponseDeliverTx.FromString,
        )
        self.CheckTx = channel.unary_unary(
            '/abci_vendor.ABCIApplication/CheckTx',
            request_serializer=vendor__pb2.RequestCheckTx.SerializeToString,
            response_deserializer=vendor__pb2.ResponseCheckTx.FromString,
        )
        self.Query = channel.unary_unary(
            '/abci_vendor.ABCIApplication/Query',
            request_serializer=vendor__pb2.RequestQuery.SerializeToString,
            response_deserializer=vendor__pb2.ResponseQuery.FromString,
        )
        self.Commit = channel.unary_unary(
            '/abci_vendor.ABCIApplication/Commit',
            request_serializer=vendor__pb2.RequestCommit.SerializeToString,
            response_deserializer=vendor__pb2.ResponseCommit.FromString,
        )
        self.InitChain = channel.unary_unary(
            '/abci_vendor.ABCIApplication/InitChain',
            request_serializer=vendor__pb2.RequestInitChain.SerializeToString,
            response_deserializer=vendor__pb2.ResponseInitChain.FromString,
        )
        self.BeginBlock = channel.unary_unary(
            '/abci_vendor.ABCIApplication/BeginBlock',
            request_serializer=vendor__pb2.RequestBeginBlock.SerializeToString,
            response_deserializer=vendor__pb2.ResponseBeginBlock.FromString,
        )
        self.EndBlock = channel.unary_unary(
            '/abci_vendor.ABCIApplication/EndBlock',
            request_serializer=vendor__pb2.RequestEndBlock.SerializeToString,
            response_deserializer=vendor__pb2.ResponseEndBlock.FromString,
        )


class ABCIApplicationServicer(object):
    """----------------------------------------
    Service Definition

    """

    def Echo(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Flush(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Info(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetOption(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeliverTx(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckTx(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Query(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Commit(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InitChain(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BeginBlock(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EndBlock(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ABCIApplicationServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Echo': grpc.unary_unary_rpc_method_handler(
            servicer.Echo,
            request_deserializer=vendor__pb2.RequestEcho.FromString,
            response_serializer=vendor__pb2.ResponseEcho.SerializeToString,
        ),
        'Flush': grpc.unary_unary_rpc_method_handler(
            servicer.Flush,
            request_deserializer=vendor__pb2.RequestFlush.FromString,
            response_serializer=vendor__pb2.ResponseFlush.SerializeToString,
        ),
        'Info': grpc.unary_unary_rpc_method_handler(
            servicer.Info,
            request_deserializer=vendor__pb2.RequestInfo.FromString,
            response_serializer=vendor__pb2.ResponseInfo.SerializeToString,
        ),
        'SetOption': grpc.unary_unary_rpc_method_handler(
            servicer.SetOption,
            request_deserializer=vendor__pb2.RequestSetOption.FromString,
            response_serializer=vendor__pb2.ResponseSetOption.SerializeToString,
        ),
        'DeliverTx': grpc.unary_unary_rpc_method_handler(
            servicer.DeliverTx,
            request_deserializer=vendor__pb2.RequestDeliverTx.FromString,
            response_serializer=vendor__pb2.ResponseDeliverTx.SerializeToString,
        ),
        'CheckTx': grpc.unary_unary_rpc_method_handler(
            servicer.CheckTx,
            request_deserializer=vendor__pb2.RequestCheckTx.FromString,
            response_serializer=vendor__pb2.ResponseCheckTx.SerializeToString,
        ),
        'Query': grpc.unary_unary_rpc_method_handler(
            servicer.Query,
            request_deserializer=vendor__pb2.RequestQuery.FromString,
            response_serializer=vendor__pb2.ResponseQuery.SerializeToString,
        ),
        'Commit': grpc.unary_unary_rpc_method_handler(
            servicer.Commit,
            request_deserializer=vendor__pb2.RequestCommit.FromString,
            response_serializer=vendor__pb2.ResponseCommit.SerializeToString,
        ),
        'InitChain': grpc.unary_unary_rpc_method_handler(
            servicer.InitChain,
            request_deserializer=vendor__pb2.RequestInitChain.FromString,
            response_serializer=vendor__pb2.ResponseInitChain.SerializeToString,
        ),
        'BeginBlock': grpc.unary_unary_rpc_method_handler(
            servicer.BeginBlock,
            request_deserializer=vendor__pb2.RequestBeginBlock.FromString,
            response_serializer=vendor__pb2.ResponseBeginBlock.SerializeToString,
        ),
        'EndBlock': grpc.unary_unary_rpc_method_handler(
            servicer.EndBlock,
            request_deserializer=vendor__pb2.RequestEndBlock.FromString,
            response_serializer=vendor__pb2.ResponseEndBlock.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'abci_vendor.ABCIApplication', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))



from forge import protos


class RpcWallet:
    __wallet_type = protos.WalletType(pk=0, hash=1, address=1)

    def __init__(self, chan):
        self.stub = protos.WalletRpcStub(chan)

    def create_wallet(
            self, wallet_type=__wallet_type, moniker='',
            passphrase='', req=None,
    ):
        """

        Parameters
        ----------
        req: RequestCreateWallet
        wallet_type: WalletType
        moniker: string
        passphrase: string

        Returns
        -------
        ResponseCreateWallet

        """
        if req is not None:
            return self.stub.create_wallet(req)
        else:
            req_kwargs = {
                'type': wallet_type,
                'moniker': moniker,
                'passphrase': passphrase,
            }
            return self.stub.create_wallet(
                protos.RequestCreateWallet(**req_kwargs),
            )

    def load_wallet(self, address='', passphrase='', req=None):
        """
        rpc call to load wallet.

        parameters
        ----------
        req: requestloadwallet
        address: string
        passphrase: string

        returns
        -------
        responseloadwallet

        """
        if req is not None:
            return self.stub.load_wallet(req)
        else:
            req_kwargs = {
                'address': address,
                'passphrase': passphrase,
            }
            return self.stub.load_wallet(
                protos.RequestLoadWallet(**req_kwargs),
            )

    def recover_wallet(
            self, passphrase='', moniker='', data=b'',
            wallet_type=__wallet_type, req=None,
    ):
        """
        rpc call to recover wallet with given passphrase.

        parameters
        ----------
        passphrase: string
        moniker: string
        req: requestrecoverwallet
        data: bytes
            data could be bytes of seed words or secret key
        wallet_type: walletType

        Returns
        -------
        ResponseRecoverWallet

        """
        if req is not None:
            return self.stub.recover_wallet(req)
        else:
            req_kwargs = {
                'data': data,
                'type': wallet_type,
                'passphrase': passphrase,
                'moniker': moniker,
            }
            req = protos.RequestRecoverWallet(**req_kwargs)
            print(req)
            return self.stub.recover_wallet(req)

    def list_wallet(self):
        """
        RPC call to list wallets

        Returns
        -------
        stream ResponseListWallets

        """
        return self.stub.list_wallet(protos.RequestListWallet())

    def remove_wallet(self, address='', req=None):
        """
        RPC call to remove wallet with given address

        Parameters
        ----------
        req: RequestRemoveWallet
        address: string

        Returns
        -------
        ResponseRemoveWallet

        """
        if req is not None:
            return self.stub.remove_wallet(req)
        else:
            req_kwargs = {
                'address': address,
            }
            return self.stub.remove_wallet(
                protos.RequestRemoveWallet(**req_kwargs),
            )

    def declare_node(self, req):
        """

        Parameters
        ----------
        req: RequestDeclareNode

        Returns
        -------
        ResponseDeclareNode

        """
        if req is not None:
            return self.stub.declare_node(req)
        else:
            return self.stub.declare_node(protos.RequestDeclareNode())



import json

from event_chain.config import config
from event_chain.utils import helpers

from forge import AbtDid
from forge.rpc import rpc as forge_rpc
from forge.utils import utils as forge_utils

APP_SK = b'0\243\016\303\017\r\305\026#~\301\227\033;\274\303pl\243 \004,' \
         b'\224c\003\261\2629&G\345\020\317w\007P\234\211g\246Q\264P\325\346' \
         b'/}E\020\304\365\216\242\033o\302\206v\203\303+\206n\212'

APP_PK = b'\317w\007P\234\211g\246Q\264P\325\346/}E\020\304\365\216\242\033o' \
         b'\302\206v\203\303+\206n\212'
APP_ADDR = "z1UT9an1Z4W1gnmzASneER2J5eqtx5jfwgx"
APP_DID_TYPE = AbtDid.parse_type_from_did(APP_ADDR)


def response_require_multisig(**kwargs):
    user_did_type = AbtDid.parse_type_from_did(kwargs.get('user_did'))

    claim = {
        'data': forge_utils.multibase_b58encode(
            user_did_type.hasher.hash(
                kwargs.get('tx').SerializeToString())),
        'meta': {
            'description': kwargs.get('description')
        },
        'method': user_did_type.hash_type,
        'origin': helpers.base58_encode_tx(kwargs.get('tx')),
        'type': "signature",
    }

    return response([claim], **kwargs)


def response_require_asset(**kwargs):
    claim = {
        'meta': {
            'description': kwargs.get('description')
        },
        'did_type': 'asset',
        'target': kwargs.get('target'),
        'type': "did",
    }

    return response([claim], **kwargs)


def response(claims, **kwargs):
    extra = {
        'url': kwargs.get('url'),
        'action': kwargs.get('action'),
        'appInfo': {
            'chainHost': 'http://{}:8211/api'.format(config.app_host),
            'chainId': forge_rpc.get_chain_info().info.network,
            'chain_token': 'TBA',
            'decimals': 16,
            'description': 'Create and join events on Event Chain',
            'icon': "http://did-workshop.arcblock.co:5000/static/images"
                    "/eventchain.png",
            'name': 'Event Chain',
            'subtitle': 'A decentralized solution for events',
        },
        'requestedClaims': claims,
        'workflow': {
            'description': kwargs.get('workflow')
        }
    }
    res = {
        'appPk': forge_utils.multibase_b58encode(APP_PK),
        'authInfo': APP_DID_TYPE.gen_and_sign(APP_SK, extra)
    }

    return json.dumps(res)



from forge import protos


class RpcEvent:

    def __init__(self, chan):
        self.stub = protos.EventRpcStub(chan)

    def subscribe(self, req=None, type=0, filter=None):
        """

        Parameters
        ----------
        req: RequestSubscribe
        type: protos.TopicType
        filter: str

        Returns
        -------
        ResponseSubscribe

        """
        if req:
            res = self.stub.subscribe(req)

        else:
            res = self.stub.subscribe(
                protos.RequestSubscribe(type=type, filter=filter),
            )
        for r in res:
            yield r

    def unsubscribe(self, topic='', req=None):
        """

        Parameters
        ----------
        topic: str
        req: RequestUnsubscribe

        Returns
        -------
        ResponseUnsubscribe

        """
        if req:
            return self.stub.unsubscribe(req)
        else:
            return self.stub.unsubscribe(
                protos.RequestUnsubscribe(topic=topic),
            )



import logging

import grpc

from forge import protos
from forge import utils
from forge.rpc.chain import RpcChain
from forge.rpc.event import RpcEvent
from forge.rpc.file import RpcFile
from forge.rpc.state import RpcState
from forge.rpc.statistic import RpcStatistic
from forge.rpc.wallet import RpcWallet


class ForgeRpc:
    __wallet_type = protos.WalletType(pk=0, hash=1, address=1)

    def __init__(self, socket):
        """
        Init Forge RPC with given socket.

        Args:
            socket: string of TCP/UDS socket
        """
        self.chan = grpc.insecure_channel(socket)
        self.logger = logging.getLogger(__name__)

        self.chain = RpcChain(self.chan)
        self.create_tx = self.chain.create_tx
        self.send_tx = self.chain.send_tx
        self.get_tx = self.chain.get_tx
        self.get_block = self.chain.get_block
        self.get_unconfirmed_tx = self.chain.get_unconfirmed_tx
        self.get_chain_info = self.chain.get_chain_info
        self.search = self.chain.search
        self.get_net_info = self.chain.get_net_info
        self.get_validators_info = self.chain.get_validators_info
        self.get_config = self.chain.get_config
        self.multisig = self.chain.multisig
        self.get_asset_address = self.chain.get_asset_address

        self.event = RpcEvent(self.chan)
        self.subscribe = self.event.subscribe
        self.unsubscribe = self.event.unsubscribe

        self.state = RpcState(self.chan)
        self.get_account_state = self.state.get_account_state
        self.get_asset_state = self.state.get_asset_state
        self.get_forge_state = self.state.get_forge_state
        self.get_stake_state = self.state.get_stake_state  # test

        self.file = RpcFile(self.chan)
        self.store_file = self.file.store_file
        self.load_file = self.file.load_file

        self.wallet = RpcWallet(self.chan)
        self.create_wallet = self.wallet.create_wallet
        self.load_wallet = self.wallet.load_wallet
        self.recover_wallet = self.wallet.recover_wallet
        self.remove_wallet = self.wallet.remove_wallet
        self.list_wallet = self.wallet.list_wallet
        self.declare_node = self.wallet.declare_node

        self.statistic = RpcStatistic(self.chan)
        self.list_asset_transactions = self.statistic.list_asset_transactions
        self.list_transactions = self.statistic.list_transactions

    def send_itx(self, type_url, itx, wallet, token, nonce=1):
        encoded_itx = utils.encode_to_any(type_url, itx)
        tx = self.create_tx(
            itx=encoded_itx, from_address=wallet.address,
            wallet=wallet, token=token, nonce=nonce
        )
        return self.send_tx(tx.tx)

    def create_asset(self, type_url, asset, wallet, token):
        encoded_asset = utils.encode_to_any(type_url, asset)
        create_asset_itx = utils.encode_to_any(
            type_url='fg:t:create_asset',
            data=protos.CreateAssetTx(data=encoded_asset),
        )
        tx = self.create_tx(
            create_asset_itx, wallet.address, wallet, token,
        )
        return self.send_tx(tx.tx)

    def update_asset(self, type_url, address, asset, wallet, token):
        encoded_asset = utils.encode_to_any(type_url, asset)
        update_asset_itx = utils.encode_to_any(
            type_url='fg:t:update_asset',
            data=protos.UpdateAssetTx(
                address=address,
                data=encoded_asset,
            ),
        )
        tx = self.create_tx(
            update_asset_itx, wallet.address, wallet, token,
        )
        return self.send_tx(tx.tx)

    def get_single_account_state(self, address):
        if address:
            accounts = self.get_account_state({'address': address})
            account = next(accounts)
            if utils.is_proto_empty(account):
                return None
            else:
                return account.state

    def get_single_tx_info(self, hash):
        if hash:
            infos = self.get_tx(hash)
            info = next(infos)
            if utils.is_proto_empty(info):
                return None
            else:
                return info.info

    def get_single_asset_state(self, address):
        if address:
            assets = self.get_asset_state({'address': address})
            asset = next(assets)
            if utils.is_proto_empty(asset):
                return None
            else:
                return asset.state

    def get_nonce(self, address):
        account = self.get_single_account_state(address)
        return account.nonce

    def multisig_consume_asset_tx(self, tx, wallet, token, data):
        return self.multisig(tx, wallet, token, data=data)



from forge import protos
from forge.config import config

stub = protos.EventRpcStub(config.get_grpc_channel())


def subscribe(req=None, type=0, filter=None):
    """

    Parameters
    ----------
    req: RequestSubscribe
    type: protos.TopicType
    filter: str

    Returns
    -------
    ResponseSubscribe

    """
    if req:
        res = stub.subscribe(req)

    else:
        res = stub.subscribe(
            protos.RequestSubscribe(type=type, filter=filter),
        )
    for r in res:
        yield r


def unsubscribe(topic='', req=None):
    """

    Parameters
    ----------
    topic: str
    req: RequestUnsubscribe

    Returns
    -------
    ResponseUnsubscribe

    """
    if req:
        return stub.unsubscribe(req)
    else:
        return stub.unsubscribe(
            protos.RequestUnsubscribe(topic=topic),
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="forge-event-chain",
    version="0.18.1",
    author="Riley Shu",
    author_email="riley@arcblock.io",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ArcBlock/forge-python-sdk",
    packages=setuptools.find_packages(
                exclude=('kvstore', 'quick_test', 'examples'),
    ),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'Flask==0.12.4',
        'Flask-AppBuilder==1.10.0',
        'Flask-Babel==0.11.1',
        'Flask-Caching==1.4.0',
        'Flask-Compress==1.4.0',
        'Flask-Login==0.2.11',
        'Flask-Migrate==2.2.1',
        'Flask-OpenID==1.2.5',
        'flask-qrcode',
        'Flask-Script==2.0.6',
        'Flask-Session==0.3.1',
        'Flask-SQLAlchemy==2.1',
        'Flask-Testing==0.7.1',
        'Flask-WTF==0.14.2',
        'flask-googlemaps',
        'forge-python-sdk==0.18.1',
        'requests',
        'click==6.7',
    ],
)



import ast
import base64
import codecs
import json
import logging
import math
from collections import Iterable
from datetime import datetime
from functools import reduce

import base58
from google.protobuf.any_pb2 import Any
from google.protobuf.internal.decoder import _DecodeVarint32
from google.protobuf.internal.encoder import _VarintBytes

from .. import protos

logger = logging.getLogger('forge-util')


def parse_to_proto(binary, proto_message):
    result = proto_message()
    result.ParseFromString(binary)
    return result


def decode_to_proto(binary, proto_message):
    data, req_len, start_pos = decode(binary)
    result = proto_message()
    result.ParseFromString(data)
    return result


def decode(binary):
    req_len, start_pos = __decode_zigzag(binary)
    res = binary[start_pos: start_pos + req_len]
    return res, req_len, start_pos


def encode(proto_data):
    size = proto_data.ByteSize()
    varint_bytes = __encode_zigzag(size)
    proto_bytes = proto_data.SerializeToString()
    return varint_bytes + proto_bytes


def encode_to_any(type_url, data):
    return Any(
        type_url=type_url,
        value=data.SerializeToString(),
    )


def __decode_zigzag(binary):
    length, start = _DecodeVarint32(binary, 0)
    res = length & 1
    if res == 1:
        return -((length + 1) >> 1), start
    elif res == 0:
        return length >> 1, start


def __encode_zigzag(integer):
    if integer >= 0:
        return _VarintBytes(integer << 1)
    else:
        return _VarintBytes(-((integer << 1) + 1))


def to_iter(to_req, data):
    if isinstance(data, dict) or isinstance(data, str):
        return iter([to_req(data)])
    elif isinstance(data, Iterable):
        return (to_req(i) for i in data)
    else:
        return iter([data])


def data_of_create_asset(tx, proto_def):
    create_tx = parse_to_proto(tx.itx.value, protos.CreateAssetTx)
    data = parse_to_proto(create_tx.data.value, proto_def)
    return data


def is_proto_empty(proto_message):
    if proto_message.SerializeToString() == b'':
        return True
    else:
        return False


def _to_bytes(n, length, endianess='big'):
    b = '%x' % n
    h = ('0' * (len(b) % 2) + b).zfill(length * 2)
    s = codecs.decode(h, "hex")
    return s if endianess == 'big' else s[::-1]


def _bytes_size(n):
    return math.ceil(len(hex(n)) / 2 - 1)


def int_to_bytes(n):
    size = int(_bytes_size(n))
    return _to_bytes(n, size)


def bytes_to_int(bytes):
    return reduce(lambda s, x: (s << 8) + x, bytearray(bytes))


def multibase_b58encode(data):
    raw = base58.b58encode(data)
    return 'z' + raw.decode()


def multibase_b58decode(data):
    if data.startswith('z'):
        return base58.b58decode((data[1:]).encode())
    raise ValueError('{} cannot be decoded by multibase'
                     ' base58.'.format(str(data)))


def to_json_b64urlsafe(dictionary):
    json_data = json.dumps(dictionary)
    return b64encode_no_padding(json_data)


def b64encode_no_padding(data):
    if isinstance(data, str):
        data = data.encode()
    return base64.urlsafe_b64encode(data).decode().rstrip('=')


def b64decode_padding_safe(data):
    if isinstance(data, str):
        data = data.encode()
    return base64.urlsafe_b64decode(
        (data + b'=' * (-len(data) % 4)))


def b64encoded_to_dict(data):
    try:
        dict_string = b64decode_padding_safe(data).decode()
        return ast.literal_eval(dict_string)
    except Exception as e:
        logger.error('Error in decoding b64urlsafe '
                     'encoded data to dictionary.')
        logger.error(e, exc_info=True)
        return {}


# TODO: utc timestamp
def current_utc_timestamp():
    return round(datetime.now().timestamp())


def clean_dict(d):
    return {k: v for k, v in d.items() if v and v != ''}
